/****************************************************************************************
*  YAROCK                                                                               *
*  Copyright (c) 2010-2013 Sebastien amardeilh <sebastien.amardeilh+yarock@gmail.com>   *
*                                                                                       *
*  This program is free software; you can redistribute it and/or modify it under        *
*  the terms of the GNU General Public License as published by the Free Software        *
*  Foundation; either version 2 of the License, or (at your option) any later           *
*  version.                                                                             *
*                                                                                       *
*  This program is distributed in the hope that it will be useful, but WITHOUT ANY      *
*  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A      *
*  PARTICULAR PURPOSE. See the GNU General Public License for more details.             *
*                                                                                       *
*  You should have received a copy of the GNU General Public License along with         *
*  this program.  If not, see <http://www.gnu.org/licenses/>.                           *
*****************************************************************************************/

#include "CollectionView/virtualplayqueue.h"
#include "utilities.h"      // pour randomInt
#include "yarocksettings.h"
#include "debug.h"

#include "MediaItem/mediaitem.h"

VirtualPlayqueue* VirtualPlayqueue::INSTANCE = 0;
/*
********************************************************************************
*                                                                              *
*    Class VirtualPlayqueue                                                    *
*                                                                              *
********************************************************************************
*/
VirtualPlayqueue::VirtualPlayqueue(QWidget *parent)
{
Q_UNUSED(parent)
    INSTANCE        = this;
}

/*******************************************************************************
    VirtualPlayqueue::clear
*******************************************************************************/
void VirtualPlayqueue::clear()
{
    foreach(MEDIA::TrackPtr track, m_tracks )
      track.reset();

    m_tracks.clear();

    m_playing_track   = MEDIA::TrackPtr(0);
    m_requested_track = MEDIA::TrackPtr(0);
}

/*******************************************************************************
    VirtualPlayqueue::appendMedia
*******************************************************************************/
void VirtualPlayqueue::appendMedia(const MEDIA::TrackPtr media)
{
    Debug::debug() << "-- VirtualPlayqueue::appendMedia";
    this->clear();
    m_tracks.append( MEDIA::TrackPtr(media) );

    setRequestedMedia(0);
    emit signal_collection_playTrack();
}

/*******************************************************************************
    VirtualPlayqueue::appendMedias
*******************************************************************************/
void VirtualPlayqueue::appendMedias(QList<MEDIA::TrackPtr> list, int startIdx)
{
    Debug::debug() << "-- VirtualPlayqueue::appendMedias";
    if (list.isEmpty()) return;

    this->clear();
    foreach(MEDIA::TrackPtr media, list)
      m_tracks.append( MEDIA::TrackPtr(media) );

    setRequestedMedia(startIdx);
    emit signal_collection_playTrack();
}

/*******************************************************************************
    VirtualPlayqueue::updatePlayingItem
*******************************************************************************/
void VirtualPlayqueue::updatePlayingItem(MEDIA::TrackPtr tk)
{
    //Debug::debug() << "VirtualPlayqueue -> updatePlayingItem";
    stopPlaying();
    if(!tk) return;

    foreach(MEDIA::TrackPtr track, m_tracks) {
      if(track == tk) {
        track->isBroken  = false;
        track->isPlaying = true;
        track->isPlayed  = true;

        m_playing_track  = track;
        break;
      }
    }
}

/*******************************************************************************
    VirtualPlayqueue::stopPlaying
*******************************************************************************/
void VirtualPlayqueue::stopPlaying()
{
    m_playing_track  = MEDIA::TrackPtr(0);
}


/*******************************************************************************
    VirtualPlayqueue::getNextMediaItem
*******************************************************************************/
MEDIA::TrackPtr VirtualPlayqueue::getNextMediaItem()
{
    //Debug::debug() << "-- VirtualPlayqueue::getNextMediaItem";
    RepeatMode  repeat_m  = (RepeatMode)SETTINGS()->_repeatMode;
    ShuffleMode shuffle_m = (ShuffleMode)SETTINGS()->_shuffleMode;

    int playingIdx = m_tracks.indexOf(m_playing_track);

    //! normal mode get next track
    if(repeat_m == RepeatOff && shuffle_m == ShuffleOff) {
      if(playingIdx < m_tracks.size()- 1) {
         setRequestedMedia(playingIdx + 1);

         return getRequestedMedia();
      }
    }

    //! repeat track mode
    if(repeat_m == RepeatTrack) {
      if(playingIdx < m_tracks.size()- 1) {
         setRequestedMedia(playingIdx);

         return getRequestedMedia();
      }
    }

    //! repeat all mode
    if(repeat_m == RepeatAll && shuffle_m == ShuffleOff) {
      if( playingIdx == m_tracks.size()- 1) {
        setRequestedMedia(0);

        return getRequestedMedia();
      }
      else if(playingIdx < m_tracks.size()- 1) {
        setRequestedMedia(playingIdx + 1);

        return getRequestedMedia();
      }
    }

    //! shuffle ==> random next until all is played
    if(shuffle_m == ShuffleOn) {
      QList<MEDIA::TrackPtr> list = unplayedMedia();
      if(list.isEmpty())
        return MEDIA::TrackPtr(0);

      int index = UTIL::randomInt(0, list.size() -1);
      MEDIA::TrackPtr media = list.at(index);

      setRequestedMedia(m_tracks.indexOf(media));

      return getRequestedMedia();
    }

    return MEDIA::TrackPtr(0);
}

/*******************************************************************************
    VirtualPlayqueue::getPrevMediaItem
*******************************************************************************/
MEDIA::TrackPtr VirtualPlayqueue::getPrevMediaItem()
{
    //Debug::debug() << "-- VirtualPlayqueue::getPrevMediaItem";
    RepeatMode  repeat_m  = (RepeatMode)SETTINGS()->_repeatMode;
    ShuffleMode shuffle_m = (ShuffleMode)SETTINGS()->_shuffleMode;

    int playingIdx = m_tracks.indexOf(m_playing_track);
    
    //! normal mode get prev track
    if(repeat_m == RepeatOff && shuffle_m == ShuffleOff) {
      if(playingIdx > 0) {
        setRequestedMedia(playingIdx -1);
        return getRequestedMedia();
      }
    }

    //! repeat track mode
    if(repeat_m == RepeatTrack) {
      if(playingIdx < m_tracks.size()- 1) {
        setRequestedMedia(playingIdx);

        return getRequestedMedia();
      }
    }

    //! repeat all mode
    if(repeat_m == RepeatAll && shuffle_m == ShuffleOff) {
      if( playingIdx == 0) {
        setRequestedMedia(m_tracks.size()- 1);

        return getRequestedMedia();
      }
      else if(playingIdx > 0) {
        setRequestedMedia(playingIdx -1);

        return getRequestedMedia();
      }
    }

    //! shuffle ==> random next until all is played
    if(shuffle_m == ShuffleOn) {
      QList<MEDIA::TrackPtr> list = unplayedMedia();
      if(list.isEmpty())
        return MEDIA::TrackPtr(0);

      int index = UTIL::randomInt(0, list.size() -1);
      MEDIA::TrackPtr media = list.at(index);

      setRequestedMedia(m_tracks.indexOf(media));

      return getRequestedMedia();
    }

    return MEDIA::TrackPtr(0);
}


/*******************************************************************************
    VirtualPlayqueue::setRequestedMediaItem
*******************************************************************************/
void VirtualPlayqueue::setRequestedMedia(int idx)
{
    if(idx >= 0 && idx < m_tracks.size())
    {
      m_requested_track = m_tracks.at(idx);
    }
    else
    {
      m_requested_track = MEDIA::TrackPtr(0);
    }
}

/*******************************************************************************
    VirtualPlayqueue::getRequestedMedia
*******************************************************************************/
MEDIA::TrackPtr VirtualPlayqueue::getRequestedMedia()
{
    return m_requested_track;
}

/*******************************************************************************
    VirtualPlayqueue::unplayedMedia
    - get the list of media not already played (use for shuffle mode)
*******************************************************************************/
QList<MEDIA::TrackPtr>  VirtualPlayqueue::unplayedMedia()
{
    QList<MEDIA::TrackPtr> list;

    foreach (MEDIA::TrackPtr media, m_tracks)
    {
        if(media->isPlayed == false)
            list << media;
    }

    return list;
}


