/****************************************************************************************
*  YAROCK                                                                               *
*  Copyright (c) 2010-2013 Sebastien amardeilh <sebastien.amardeilh+yarock@gmail.com>   *
*                                                                                       *
*  This program is free software; you can redistribute it and/or modify it under        *
*  the terms of the GNU General Public License as published by the Free Software        *
*  Foundation; either version 2 of the License, or (at your option) any later           *
*  version.                                                                             *
*                                                                                       *
*  This program is distributed in the hope that it will be useful, but WITHOUT ANY      *
*  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A      *
*  PARTICULAR PURPOSE. See the GNU General Public License for more details.             *
*                                                                                       *
*  You should have received a copy of the GNU General Public License along with         *
*  this program.  If not, see <http://www.gnu.org/licenses/>.                           *
*****************************************************************************************/

// local
#include "engineplayer.h"
#include "utilities.h"
#include "yarocksettings.h"
#include "MediaItem/mediaitem.h"
#include "debug.h"


// Qt
#include <QFileInfo>
#include <QList>

// phonon
#include <phonon/backendcapabilities.h>
#include <phonon/effect.h>
#include <phonon/effectparameter.h>


#include <cmath>


EnginePlayer* EnginePlayer::INSTANCE = 0;


/*
********************************************************************************
*                                                                              *
*    Class EnginePlayer                                                        *
*                                                                              *
********************************************************************************
*/

EnginePlayer::EnginePlayer()
{
    INSTANCE = this;
    Debug::debug() << "engine player init";

    _mediaObject = new Phonon::MediaObject(this);
    _audioOutput = new Phonon::AudioOutput(Phonon::MusicCategory, this);

    // By default tick every 1 second
    _mediaObject->setTickInterval(100);

    // Get the next track when there is 2 seconds left on the current one
    // in case of playing track from track view
    _mediaObject->setPrefinishMark( 2000 );
    _mediaObject->setTransitionTime(100); //! GAPLESS/CROSSFADE

    connect(_mediaObject,SIGNAL(finished()),this,SLOT(slot_queue_finished()));
    connect(_mediaObject,SIGNAL(aboutToFinish()),this,SLOT(slot_media_aboutToFinish()));

    connect(_mediaObject,SIGNAL(stateChanged(Phonon::State,Phonon::State)),this,SLOT(slotStateChanged(Phonon::State,Phonon::State)),Qt::DirectConnection);

    connect(_mediaObject,SIGNAL(tick(qint64)),this,SIGNAL(trackPositionChanged(qint64)));
    connect(_mediaObject,SIGNAL(totalTimeChanged(qint64)),this,SLOT(slotTrackTotalTimeChanged(qint64)));

    connect(_mediaObject,SIGNAL(currentSourceChanged( const Phonon::MediaSource & )),this,SLOT(slotTrackChanged( const Phonon::MediaSource & )));

    connect(_mediaObject, SIGNAL(metaDataChanged()), this,SLOT(slotMetaDataChanged()));

    connect( _audioOutput, SIGNAL( volumeChanged( qreal ) ),this, SLOT( slotVolumeChanged( qreal ) ) );
    connect( _audioOutput, SIGNAL( mutedChanged( bool ) ), this,  SLOT( slotMutedChanged( bool ) ) );


    _phononPath     = Phonon::createPath(_mediaObject,_audioOutput);
    _audioEqualizer = 0;

    //! volume init
    _playerVolume = qBound<qreal>( 0, qRound(_audioOutput->volume()*100), 100 );

    _engineState = ENGINE::STOPPED;
    _oldState    = ENGINE::STOPPED;

    m_currentMediaItem = MEDIA::TrackPtr(0);
    m_nextMediaItem    = MEDIA::TrackPtr(0);
    m_playing_url      = QString();

    //! only create pre-amp if we have replaygain on, VolumeFaderEffect can cause phonon issues
    if( SETTINGS()->_replaygain != SETTING::ReplayGainOff )
    {
      _preamp = new Phonon::VolumeFaderEffect( this );
      _phononPath.insertEffect( _preamp );
    }
    else {
      _preamp = 0;
    }

    //! add an equalizer effect if available
    QList<Phonon::EffectDescription> mEffectDescriptions = Phonon::BackendCapabilities::availableAudioEffects();
    foreach ( const Phonon::EffectDescription &mDescr, mEffectDescriptions )
    {
      if ( mDescr.name() == QLatin1String( "KEqualizer" ) )
      {
          Debug::debug() << "EnginePlayer::create equalizer";
          _audioEqualizer = new Phonon::Effect( mDescr, this );

          if( SETTINGS()->_enableEq ) {
            addEqualizer();
            loadEqualizerSettings();
          }
      }
    }
}

EnginePlayer::~EnginePlayer()
{
    _mediaObject->blockSignals(true);
    _mediaObject->stop();
    delete _mediaObject;
    delete _audioOutput;
}


/*******************************************************************************
    Playing method
*******************************************************************************/
void EnginePlayer::play()
{
    _mediaObject->play();
}

void EnginePlayer::pause()
{
    _mediaObject->pause();
}

void EnginePlayer::stop()
{
   _mediaObject->stop();
   _mediaObject->clearQueue();

   m_playing_url.clear();
   if(m_currentMediaItem) {
     m_currentMediaItem.reset();
     m_currentMediaItem = MEDIA::TrackPtr(0);
   }

   if(m_nextMediaItem) {
     m_nextMediaItem.reset();
     m_nextMediaItem    = MEDIA::TrackPtr(0);
   }

   if(_engineState != ENGINE::STOPPED) {
     _engineState =  ENGINE::STOPPED;
     _oldState    =  ENGINE::STOPPED;
     emit engineStateChanged();
   }

}


/*******************************************************************************
    Phonon state
*******************************************************************************/
void EnginePlayer::slotStateChanged(Phonon::State newState, Phonon::State oldState)
{
Q_UNUSED(oldState)
    //Debug::warning() << "EnginePlayer::slotStateChanged : " << newState ;
    //Debug::warning() << "EnginePlayer::m_playing_url : " << m_playing_url ;
    if(newState == oldState)
      return;

    switch (newState) {
      /**************** LOADING **********************/
      case Phonon::LoadingState:
         _engineState = ENGINE::STOPPED;
      break;
      /**************** STOPPED **********************/
      case Phonon::StoppedState:
         _engineState = ENGINE::STOPPED;
      break;
      /**************** PLAYING **********************/
     case Phonon::PlayingState:
        _engineState = ENGINE::PLAYING;
      break;
      /**************** BUFFERING ********************/
      case Phonon::BufferingState:
         //_engineState = ENGINE::STOPPED;
      break;
      /**************** PAUSE ************************/
      case Phonon::PausedState:
         _engineState = ENGINE::PAUSED;
      break;
      /**************** ERROR ************************/
      case Phonon::ErrorState:
        _engineState = ENGINE::ERROR;
      break;
      default:break;
    }

    if(_oldState != _engineState)
    {
      if(_engineState != ENGINE::STOPPED)
      {
        Debug::debug() << "EnginePlayer::engineStateChanged :" << stateToString(_engineState);
        emit engineStateChanged();
      }
      else if(currentMediaSource().isEmpty())
      {
        Debug::debug() << "EnginePlayer::engineStateChanged :" << stateToString(_engineState);
        emit engineStateChanged();
      }

      _oldState = _engineState;
    }
}


/*******************************************************************************
    current playing Item info
*******************************************************************************/
void  EnginePlayer::slotTrackTotalTimeChanged(qint64 total_time_ms)
{
    //Debug::debug() << "EnginePlayer::slotTrackTotalTimeChanged" ;

    //! Gstreamer send totaltimechange before current source change
    //! It seems to be not the case for phonon-vlc
    if(m_nextMediaItem && m_nextMediaItem->type() == TYPE_TRACK)
    {
      return;
    }

    if(m_currentMediaItem && m_currentMediaItem->type() == TYPE_TRACK)
    {
      int duration = MEDIA::TrackPtr::staticCast(m_currentMediaItem)->duration;
      if( duration > 0 )
      {
        emit trackTotalTimeChanged( duration * 1000 );
      }
      else
      {
        emit trackTotalTimeChanged( _mediaObject->totalTime() ); // milliseconds
      }
    }
    else
    {
      emit trackTotalTimeChanged(total_time_ms);
    }
}


qint64 EnginePlayer::enginePlayingTrackLength()
{
   if(m_currentMediaItem && m_currentMediaItem->type() == TYPE_TRACK)
   {
      int duration = m_currentMediaItem->duration;

      if( duration > 0 )
      {
        return duration * 1000;
      }
      else
      {
        return _mediaObject->totalTime();
      }
    }

    return _mediaObject->totalTime();
}



// private
QString EnginePlayer::currentMediaSource()
{
    if (_mediaObject->currentSource().type() == Phonon::MediaSource::LocalFile)
      return _mediaObject->currentSource().url().toLocalFile();
    else
      return _mediaObject->currentSource().url().toString();

    return QString();
}


static const qreal log10over20 = 0.1151292546497022842; // ln(10) / 20

void EnginePlayer::slotTrackChanged( const Phonon::MediaSource &source )
{
    Debug::debug() << "EnginePlayer::slotTrackChanged";

    // send url to string or toLocalfile
    // do not emit for empty source -> player stop is assume by next process
    m_playing_url = currentMediaSource();

    if( source.type() == Phonon::MediaSource::Empty )
    {
        Debug::debug() << "EnginePlayer::slotTrackChanged Empty MediaSource" ;
        return;
    }

    if(m_nextMediaItem) {
      Debug::debug() << "EnginePlayer -> slotTrackChanged next media item from queue";
      m_currentMediaItem = MEDIA::TrackPtr(m_nextMediaItem);
      m_nextMediaItem.reset();
      m_nextMediaItem    = MEDIA::TrackPtr(0);
    }

    //Debug::debug() << "EnginePlayer::slotTrackChanged #2";

    if(!m_currentMediaItem)
    {
        Debug::error() << "EnginePlayer::no media set";
        return;
    }
    else if ( (m_currentMediaItem->type() == TYPE_TRACK) &&
              (SETTINGS()->_replaygain != SETTING::ReplayGainOff ) )
    {
        //Debug::debug() << "EnginePlayer::slotTrackChanged #3";

        if( !_preamp )  {
            _preamp = new Phonon::VolumeFaderEffect( this );
            _phononPath.insertEffect( _preamp );
        }

        qreal gain, peak;
        if ( SETTINGS()->_replaygain != SETTING::AlbumReplayGain ) {
          gain =  m_currentMediaItem->trackGain;
          peak =  m_currentMediaItem->trackPeak;
        }
        else {
          gain =  m_currentMediaItem->albumGain;
          peak =  m_currentMediaItem->albumPeak;
        }

        if ( gain + peak > 0.0 )
        {
            Debug::debug() << "EnginePlayer -> slotTrackChanged Gain of" << gain << "would clip at absolute peak of" << gain + peak;
            gain -= gain + peak;
        }
        Debug::debug() << "EnginePlayer -> slotTrackChanged Using gain of" << gain << "with relative peak of" << peak;
        // we calculate the volume change ourselves, because m_preamp->setVolumeDecibel is
        // a little confused about minus signs
        _preamp->setVolume( exp( gain * log10over20 ) );
        _preamp->fadeTo( exp( gain * log10over20 ), 0 );
    }
    else if( _preamp )
    {
        //Debug::debug() << "EnginePlayer::reset preamp";
        _preamp->setVolume( 1.0 );
        _preamp->fadeTo( 1.0, 0 );
    }

    emit trackChanged( m_playing_url );
    emit playingInfoChanged();
}

void EnginePlayer::slotMetaDataChanged()
{
    //Debug::debug() << "### EnginePlayer::slotMetaDataChanged ";

    if(m_currentMediaItem->type() != TYPE_STREAM) return;

    const QMap<QString, QString> &metaData = _mediaObject->metaData();

    m_currentMediaItem->title  = (metaData.value("TITLE") != "Streaming Data") ? metaData.value("TITLE") : "";
    m_currentMediaItem->album  = (metaData.value("ALBUM") != "Streaming Data") ? metaData.value("ALBUM") : "";
    m_currentMediaItem->artist = (metaData.value("ARTIST") != "Streaming Data") ? metaData.value("ARTIST") : "";

    //Debug::debug() << "EnginePlayer::slotMetaDataChanged title: " << m_currentMediaItem->title;
    //Debug::debug() << "EnginePlayer::slotMetaDataChanged album: " << m_currentMediaItem->album;
    //Debug::debug() << "EnginePlayer::slotMetaDataChanged artist: " << m_currentMediaItem->artist;

    emit streamInfoChanged();
}

/*******************************************************************************
    Track ended
*******************************************************************************/
void EnginePlayer::slot_media_aboutToFinish()
{
    Debug::debug() << "EnginePlayer -> slot_media_aboutToFinish";

    if( m_currentMediaItem && m_currentMediaItem->isStopAfter )
    {
      emit player_stop_requested();
      return;
    }

    if(!m_nextMediaItem)
      emit trackAboutToFinish(); // needed to set next media item
}


void  EnginePlayer::slot_queue_finished()
{
    //! FIXME VLC phonon backend not emit about to finish every time (source is not taken from the queue)
    Debug::debug() << "EnginePlayer::slot_queue_finished";
    if(m_nextMediaItem)
    {
      Debug::debug() << "EnginePlayer::slot_queue_finished next mediaitem present !!";
      setMediaItem(m_nextMediaItem);
    }
    else if( _mediaObject->queue().isEmpty())
    {
      Debug::debug() << "EnginePlayer::slot_queue_finished QUEUE IS EMPTY emit queueFinished";
      emit queueFinished();
    }

    /*
    //! DEBUG check queue
    Debug::debug() << "EnginePlayer _mediaObject->queue().size() :" << _mediaObject->queue().size() ;
    foreach (Phonon::MediaSource source, _mediaObject->queue() )
    {
     Debug::debug() << "EnginePlayer #queue -> url :" << source.url() ;
    }
    */
}

/*******************************************************************************
    Set New source and play it !
*******************************************************************************/
void EnginePlayer::setMediaItem(MEDIA::TrackPtr track)
{
    Debug::debug() << "EnginePlayer::setMediaItem";

    _mediaObject->blockSignals(true);
    _mediaObject->stop();
    _mediaObject->clearQueue(); // AJOUT
    m_playing_url.clear();
    _mediaObject->blockSignals(false);

    if(m_currentMediaItem) {
      m_currentMediaItem.reset();
      m_currentMediaItem = MEDIA::TrackPtr(0);
    }

    m_currentMediaItem = MEDIA::TrackPtr(track);
    //Debug::debug() << "EnginePlayer -> setMediaItem mi url:" << mi->data.url;

    if(m_nextMediaItem) {
      m_nextMediaItem.reset();
      m_nextMediaItem    = MEDIA::TrackPtr(0);
    }

    //! BEGIN crash on broken path with vlc backend
    if( m_currentMediaItem->isBroken )
    {
      //TRACK::isBroken -> works for TRACK or STREAM
      Debug::warning() << "EnginePlayer:: Track path seems to be broken:" << m_currentMediaItem->url;
      _mediaObject->blockSignals(false);
      stop();
      return;
    }
    //!END

    const QString path = MEDIA::Track::path(track->url);
    _mediaObject->setCurrentSource( Phonon::MediaSource(path) );

    _mediaObject->play();
}


/*******************************************************************************
    MediaObject QUEUE
*******************************************************************************/
void EnginePlayer::setNextMediaItem(MEDIA::TrackPtr track)
{
    Debug::debug() << "EnginePlayer::setNextMediaItem";

    if(m_nextMediaItem)
      m_nextMediaItem.reset();

    m_nextMediaItem = MEDIA::TrackPtr(track);

    const QString path = track->url;

    if( MEDIA::isLocal(path) )
      _mediaObject->enqueue( Phonon::MediaSource(QFileInfo(path).filePath()) );
    else
      _mediaObject->enqueue( Phonon::MediaSource(QUrl(path)) );


    /*
    //! DEBUG check queue
    foreach (Phonon::MediaSource source, _mediaObject->queue() )
    {
     Debug::debug() << "EnginePlayer #queue -> url :" << source.url() ;
    }
   */
}


/*******************************************************************************
    MediaObject Seek
*******************************************************************************/
void EnginePlayer::seek( int ms )
{
    if( _mediaObject->isSeekable() )
    {
        Debug::debug() << "seek to: " << ms;

        _mediaObject->seek( ms );
        emit trackPositionChanged( ms );
    }
    else
        Debug::warning() << "not seekable media";
}

/*******************************************************************************
    Equalizer method
*******************************************************************************/
bool EnginePlayer::isEqualizerAvailable()
{
    QList<Phonon::EffectDescription> effects = Phonon::BackendCapabilities::availableAudioEffects();

    foreach (Phonon::EffectDescription effect, effects) {
      if(effect.name()=="KEqualizer")
        return true;
     }
     return false;
}

void EnginePlayer::addEqualizer()
{
/*  QList<Phonon::EffectDescription> descriptions = Phonon::BackendCapabilities::availableAudioEffects();

  foreach (Phonon::EffectDescription description, descriptions)  {
    if(description.name() == "KEqualizer")
    {
      _audioEqualizer = new Phonon::Effect(description, this);
      _phononPath.insertEffect(_audioEqualizer);
    }
  }
  */
    if(!_audioEqualizer) {
      Debug::debug() << " EnginePlayer::addEqualizer no equalizer available !";
      return;
    }

    // Insert effect into path if needed
    if( _phononPath.effects().indexOf( _audioEqualizer ) == -1 )
    {
      if( !_phononPath.effects().isEmpty() )
      {
        _phononPath.insertEffect( _audioEqualizer, _phononPath.effects().first() );
      }
      else
      {
        _phononPath.insertEffect( _audioEqualizer );
      }
    }
}


void EnginePlayer::removeEqualizer()
{
    //Debug::debug() << " EnginePlayer::removeEqualizer !";
    if( _phononPath.effects().indexOf( _audioEqualizer ) != -1 ) {
      _phononPath.removeEffect( _audioEqualizer );
    }
}


void EnginePlayer::applyEqualizer(QList<int> gains)
{
    if(!_audioEqualizer) {
      Debug::debug() << "EnginePlayer --> no equalizer initialized !!";
      return;
    }

    QList<Phonon::EffectParameter> effect_param_list = _audioEqualizer->parameters();

    QListIterator<int> it_effect_param( gains );
    double scaled_value;
    foreach( const Phonon::EffectParameter &effect_param, effect_param_list )
    {
       scaled_value = it_effect_param.hasNext() ? it_effect_param.next() : 0;
       scaled_value *= ( qAbs(effect_param.maximumValue().toDouble() ) +  qAbs( effect_param.minimumValue().toDouble() ) );
       scaled_value /= 200.0;
       //Debug::debug() << "EnginePlayer set equalizer value :" << scaled_value;
       _audioEqualizer->setParameterValue( effect_param, scaled_value );
    }
}

void EnginePlayer::loadEqualizerSettings()
{
    const QString preset_name = SETTINGS()->_currentPreset;

    if( SETTINGS()->_presetEq.keys().contains(preset_name) ) {
      QList<int> gains;

      gains << SETTINGS()->_presetEq[preset_name].preamp;
      for (int i=0 ; i<Equalizer::kBands ; ++i)
        gains << SETTINGS()->_presetEq[preset_name].gain[i];

      applyEqualizer(gains);
    }
}

/*******************************************************************************
    volume method
*******************************************************************************/
void EnginePlayer::slotVolumeChanged( qreal newVolume )
{
    int percent = qBound<qreal>( 0, qRound(newVolume * 100), 100 );

    if ( _playerVolume != percent ) {
      _playerVolume = percent;
      emit volumeChanged( percent );
    }
}

int EnginePlayer::setVolume( int percent )
{
    percent = qBound<qreal>( 0, percent, 100 );
    _playerVolume = percent;

    const qreal volume =  percent / 100.0;

    if (_audioOutput->volume() != volume )  {
        _audioOutput->setVolume( volume );
        emit volumeChanged( percent );
    }

    return percent;
}

void EnginePlayer::slotMutedChanged( bool mute )
{
    emit muteStateChanged( mute );
}

void EnginePlayer::setMuted( bool mute )
{
    _audioOutput->setMuted( mute );
    if( !isMuted() )
        setVolume( _playerVolume );

    emit muteStateChanged( mute );
}

/*******************************************************************************
    Enggine state to String
*******************************************************************************/
QString EnginePlayer::stateToString(ENGINE::E_ENGINE_STATE state)
{
    switch(state)
    {
      case ENGINE::PLAYING: return QString("PLAYING");break;
      case ENGINE::PAUSED:  return QString("PAUSED");break;
      case ENGINE::STOPPED: return QString("STOPPED");break;
      case ENGINE::ERROR:   return QString("ERROR");break;
      default : return QString("unknown state");
    }
}
